// cml2efc.cpp

#include "stdafx.h"
#include <string.h>

///////////////////////////////////////////////////////////////////////
// Declarations for Decode routines by John S. Fine
//

struct IR {
				  // General information about the whole decode being attempted
	FILE *ftxt;   // Output file
	FILE *fcml;   // Input file
	long freq;    // Frequency
	int h_once;
	int h_again;
	int once;     // Number of pairs in the send-once portion
	int again;    // Number of pairs in the repeat section
	int found;
	int data[1026];

	void ProcessCML();
	void ProcessSignal();
	void showData();
	void decoder();
	int TryDLL();
	int TryDLL_once(unsigned int iaContext[2]);
};


///////////////////////////////////////////////////////////////////////
// Declarations for DLL Decode routines by Daniel Arnold
//
typedef void (CALLBACK* typeDecodeIR) (unsigned int* Context,
		const int* TpaiBursts, 	
		const int TiFreq, 
		const int TiSingleBurstCount, 
		const int TiRepeatBurstCount, 
		char* TsProtocol, 
		int* TiDevice, 
		int* TiSubDevice, 
		int* TiOBC, 
		int TaiHex[4], 
		char* TsMisc, 
		char* TsError);

typedef int (CALLBACK* typeProtocolSupportLevel) (char * TsProtocol);

int HexCmdToEFC(int iHexCmd);
void LoadDll();
int ProtocolSupportLevel(char *Protocol);

HINSTANCE hDll = NULL;
typeDecodeIR lpfnDecodeIR;
typeProtocolSupportLevel lpfnProtocolSupportLevel;

//
///////////////////////////////////////////////////////////////////////

void ProcessCML(char * fn);

int main(int argc, char* argv[])
{
	// open the cml and decode files
	
	char fn1[256];
	char fn2[256];

	if (argc!=2) {
		fprintf(stderr,"use: cml2efc filename\n");
		return 1;
	}
	strcpy(fn1,argv[1]);
	char * p = strrchr(fn1,'.');
	if (p==NULL)
		strcat(fn1,".cml");

	strcpy(fn2,argv[1]);
	p = strrchr(fn2,'.');
	if (p!=NULL)
		*p = 0;
	strcat(fn2,".txt");
	
	IR ir;

	ir.fcml = fopen(fn1,"rb");
	if (!ir.fcml)
		exit(-1);
	ir.ftxt = fopen(fn2,"w");
	if (!ir.ftxt)
		exit(-1);
	
	//Load the DecodeIR dll
	LoadDll();

	ir.ProcessCML();
	fclose(ir.ftxt);
	fclose(ir.fcml);

	//Unload the DecodeIR dll if it was loaded
	if (hDll != 0)
		FreeLibrary(hDll);

	return 0;
}


void IR::ProcessCML()
{
	int c=0;
	while (!feof(fcml))
	{
		if (   (c==0xBB || (c=fgetc(fcml)) == 0xBB)
		    && (c=fgetc(fcml)) == 0xBB
		    && (c=fgetc(fcml)) == 0xBB
		    && (c=fgetc(fcml)) == 0xBB )
		{
			char trash[12];
			fread(trash, 1, 12, fcml);

			static char manu[22];
			fread(manu, 1, 21, fcml);
			fprintf(ftxt,"%s",manu);

			static char kind[22];
			fread(kind, 1, 21, fcml);
			fprintf(ftxt," <%s> ",kind);

			static char name[22];
			fread(name, 1, 21, fcml);
			fprintf(ftxt,"%s\n",name);

			c=0;
		}
		if (    c==0xCC
			&& (c=fgetc(fcml)) == 0xCC
			&& (c=fgetc(fcml)) == 0xCC
			&& (c=fgetc(fcml)) == 0xCC )
		{
			ProcessSignal();
		}
	}
	fprintf(ftxt,"\n");
}

void IR::ProcessSignal()
{
	unsigned char x[23];
	char n[22];
	unsigned char tm[256][3];
	unsigned char raw[1024];
	int time[256];
	fread(x, 1, 23, fcml);
	fread(n, 1, 21, fcml);
	n[21]=0;

	int nl = strlen(n);
	n[nl] = '\t';
	if (nl<16)
	{
		if (nl<8)
			n[++nl] = '\t';
		n[++nl] = '\t';
	}
	n[1+nl] = 0;
	fprintf(ftxt,"%s",n);

	fread(tm, 3, x[11], fcml);
	while (x[11])
	{
		unsigned char* v=tm[--x[11]];
		time[x[11]] = (v[2]+1)*32768-v[0]/2-128*v[1];
	};
	h_once = x[12] + 256*x[13];
	h_again = x[14]+ 256*x[15];
	unsigned total = h_once + h_again;
	if (total <= 1024)
	{
		fread(raw, 1, total, fcml);
		while (total)
		{
			--total;
			data[total] = time[ raw[total] - 1 ];
		}
	}

	int wav = x[9]+256*x[10];
	freq = 0;
	if (wav)
		freq = 1000000000 / wav;
	TryDLL();
	showData();
	fprintf(ftxt, "\n");

}

void IR::showData()
{
	fprintf(ftxt,"[%d]", freq);
	int total=h_once+h_again;
	for (int ndx=0; ndx<total; ++ndx)
	{
		if (ndx==h_once)
			fprintf(ftxt,";");
		if (ndx&1)
			fprintf(ftxt,"-");
		else if (ndx&31)
			fprintf(ftxt," ");
		else if (ndx)
			fprintf(ftxt,"\n");
		fprintf(ftxt,"%d",data[ndx]);
	}
	fprintf(ftxt,"\n");
}

// DLL Decode routines
void LoadDll() 
{
	hDll = LoadLibrary("DecodeIR.dll");
	if (hDll != NULL)
	{
		lpfnDecodeIR = (typeDecodeIR)GetProcAddress(hDll,
			"DecodeIR");
		if (lpfnDecodeIR==0) {
			FreeLibrary(hDll);
			hDll = NULL;
		};

		lpfnProtocolSupportLevel = (typeProtocolSupportLevel)GetProcAddress(hDll,
			"ProtocolSupportLevel");
		if (lpfnProtocolSupportLevel==0) {
			FreeLibrary(hDll);
			hDll = NULL;
		};
	};
}

int ProtocolSupportLevel(char *Protocol) {
	if (hDll != NULL)
		return lpfnProtocolSupportLevel(Protocol);
	else
		return 0;

}

int HexCmdToEFC(int iHexCmd) {

	int Result;
	Result = ( ( ((iHexCmd << 3)+(iHexCmd >> 5)) ^ 0xAE ) + 100 ) & 0xFF;
	return Result;
}

int IR::TryDLL_once(unsigned int iaContext[2])
{
	char szProtocol[255] = "";
	int iDevice = -1;
	int iSubDevice= -1;
	int iOBC= -1;
	int lpiHex[4] = {-1,-1,-1,-1};
	char szMisc[255]= "";
	char szError[255]= "";

	//Call the DecodeIR function
	lpfnDecodeIR(iaContext, data, freq, once, again, szProtocol,&iDevice,&iSubDevice,&iOBC,lpiHex,szMisc,szError);

	//Print the result
	if (szProtocol[0])  {
		if (found)
			fprintf(ftxt,"\t\t\t");
		found = 1;
        fprintf(ftxt, "(%s):", szProtocol);
        if (iDevice != -1)
            fprintf(ftxt, "%d", iDevice);
        if (iSubDevice != -1)
            fprintf(ftxt, ".%d", iSubDevice);
        if (iOBC != -1)
            fprintf(ftxt, ":%d", iOBC);
        if (lpiHex[0] != -1)
        {
            fprintf(ftxt, " EFC=%03d", HexCmdToEFC(lpiHex[0]) );
            for (int ndx=1; ndx<4 && lpiHex[ndx]!=-1; ++ndx)
                fprintf(ftxt, " or %03d", HexCmdToEFC(lpiHex[ndx]) );
        }
        if (szMisc[0])
            fprintf(ftxt, " {%s}", szMisc);
        fprintf(ftxt, "\n");
		
		return 1;
	}
	
	return 0;
}

int IR::TryDLL()
{
	unsigned int iaContext[2] = {0,0};
	found = 0;

	once = h_once/2;
	if (h_once & 1)
	{
		// Don't really know what it means for this to be odd
		// but make a pluasible patch
		if (h_again)
			data[h_once+h_again] = data[h_once];
		else
			data[h_once] = 999999;
		once++;
	}
	again = h_again/2;
	if (h_again & 1)
	{
		// This really makes no sense, but I've seen samples of it
		// This is the best kludge I could think of
		data[once*2+h_again] = 999999;
		again++;
	}

    while ( TryDLL_once(iaContext) )
		;
	
	return found;
}
